#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""

"""
from binaryornot.check import is_binary
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO

import struct
import dessia_common as dc
import volmdlr as vm
import volmdlr.faces as vmf



class Stl(dc.DessiaObject):
    """STL files are used to represent simple 3D models, defined using
    triangular 3D faces.
    
    Initially it was introduced as native format for 3D Systems
    Stereolithography CAD system, but due to its extreme simplicity, it
    was adopted by a wide range of 3D modelling, CAD, rapid prototyping
    and 3D printing applications as the simplest 3D model exchange
    format.
    
    STL is extremely bare-bones format: there are no complex headers, no
    texture / color support, no units specifications, no distinct vertex
    arrays. Whole model is specified as a collection of triangular
    faces.
    
    There are two versions of the format (text and binary), this spec
    describes binary version.
    """
    def __init__(self, triangles, name=''):
        self.triangles = triangles
        self.name = name
        
    @classmethod
    def points_from_file(cls, filename:str, distance_multiplier=0.001):
        if is_binary(filename):       
            with open(filename, 'rb') as file:
                stream = KaitaiStream(file)
                name = stream.read_bytes(80).decode('utf8')
                num_triangles = stream.read_u4le()
                
                all_points = []
                for i in range(num_triangles):
                    if i % 5000 == 0:
                        print('reading stl', round(i/num_triangles*100, 2), '%')
                    normal = vm.Vector3D(stream.read_f4le(), stream.read_f4le(), stream.read_f4le())
                    p1 = vm.Point3D(distance_multiplier*stream.read_f4le(),
                                    distance_multiplier*stream.read_f4le(),
                                    distance_multiplier*stream.read_f4le())
                    p2 = vm.Point3D(distance_multiplier*stream.read_f4le(),
                                    distance_multiplier*stream.read_f4le(), 
                                    distance_multiplier*stream.read_f4le())
                    p3 = vm.Point3D(distance_multiplier*stream.read_f4le(),
                                    distance_multiplier*stream.read_f4le(),
                                    distance_multiplier*stream.read_f4le())
                    all_points.extend([p1, p2, p3])                        
        
                    stream.read_u2le()
        return all_points

    @classmethod
    def from_file(cls, filename:str, distance_multiplier=0.001):
        if is_binary(filename):       
            with open(filename, 'rb') as file:
                stream = KaitaiStream(file)
                name = stream.read_bytes(80).decode('utf8')
                # print(name)
                num_triangles = stream.read_u4le()
                # print(num_triangles)
                
                triangles = [None] * (num_triangles)
                invalid_triangles = []
                for i in range(num_triangles):
                    if i % 5000 == 0:
                        print('reading stl', round(i/num_triangles*100, 2), '%')
                    normal = vm.Vector3D(stream.read_f4le(), stream.read_f4le(), stream.read_f4le())
                    p1 = vm.Point3D(distance_multiplier*stream.read_f4le(),
                                    distance_multiplier*stream.read_f4le(),
                                    distance_multiplier*stream.read_f4le())
                    p2 = vm.Point3D(distance_multiplier*stream.read_f4le(),
                                    distance_multiplier*stream.read_f4le(), 
                                    distance_multiplier*stream.read_f4le())
                    p3 = vm.Point3D(distance_multiplier*stream.read_f4le(),
                                    distance_multiplier*stream.read_f4le(),
                                    distance_multiplier*stream.read_f4le())
                    try : 
                        triangles[i] = vmf.Triangle3D(p1, p2, p3)
                    except ZeroDivisionError :
                        invalid_triangles.append(i)
                        
        
                    stream.read_u2le()
                    # print(abr)
            if invalid_triangles :
                print('invalid_triangles number: ', len(invalid_triangles))
                for i in invalid_triangles[::-1] :
                    del triangles[i]
        else:
            with open(filename, 'r') as file:
                header = file.readline()
                name = header[6:]
                triangles = []
                points = []
                for line in file.readlines():
                    if 'vertex' in line:
                        line = line.replace('vertex', '')
                        line = line.lstrip(' ')
                        x, y, z = line.split(' ')
                        points.append(vm.Point3D(distance_multiplier*float(x), 
                                                 distance_multiplier*float(y), 
                                                 distance_multiplier*float(z)))
                    if 'endfacet' in line:
                        try: 
                            triangles.append(vmf.Triangle3D(*points))
                        except ZeroDivisionError :
                            pass
                        points = []
        
        return cls(triangles, name=name)
    
    @classmethod
    def from_file_points(cls, filename:str, distance_multiplier=0.001):
        all_points = []
        if is_binary(filename):       
            with open(filename, 'rb') as file:
                stream = KaitaiStream(file)
                name = stream.read_bytes(80).decode('utf8')
                # print(name)
                num_triangles = stream.read_u4le()
                # print(num_triangles)
                all_points = []
                for i in range(num_triangles):
                    if i % 5000 == 0:
                        print('reading stl', round(i/num_triangles*100, 2), '%')
                    normal = vm.Vector3D(stream.read_f4le(), stream.read_f4le(), stream.read_f4le())
                    # print(n)
                    p1 = vm.Point3D(distance_multiplier*stream.read_f4le(),
                                    distance_multiplier*stream.read_f4le(),
                                    distance_multiplier*stream.read_f4le())
                    p2 = vm.Point3D(distance_multiplier*stream.read_f4le(),
                                    distance_multiplier*stream.read_f4le(), 
                                    distance_multiplier*stream.read_f4le())
                    p3 = vm.Point3D(distance_multiplier*stream.read_f4le(),
                                    distance_multiplier*stream.read_f4le(),
                                    distance_multiplier*stream.read_f4le())
                    # print(p1, p2, p3)
                    all_points.extend([p1, p2, p3])                        
                    stream.read_u2le()
        print('all points :', len(all_points))
        return all_points

    def save_to_binary_file(self, filepath, distance_multiplier=1000):
        BINARY_HEADER ="80sI"
        BINARY_FACET = "12fH"
        if not filepath.endswith('.stl'):
            filepath += '.stl'
            print('Adding .stl extension: ', filepath)

        with open(filepath, 'wb') as file:
            file.seek(0)
            # counter = 0
            file.write(struct.pack(BINARY_HEADER, self.name.encode('utf8'), len(self.triangles)))
            # counter += 1
            for triangle in self.triangles:
                data = [
                    0., 0., 0.,
                    distance_multiplier*triangle.point1.x, distance_multiplier*triangle.point1.y, distance_multiplier*triangle.point1.z,
                    distance_multiplier*triangle.point2.x, distance_multiplier*triangle.point2.y, distance_multiplier*triangle.point2.z,
                    distance_multiplier*triangle.point3.x, distance_multiplier*triangle.point3.y, distance_multiplier*triangle.point3.z,
                    0
                    ]
                file.write(struct.pack(BINARY_FACET, *data))
            file.close()
            
    def to_closed_shell(self):
        return vmf.ClosedShell3D(self.triangles, name=self.name)
    
    def to_open_shell(self):
        return vmf.OpenShell3D(self.triangles, name=self.name)

    
    def extract_points(self):
            
        points1 = [t.point1 for t in self.triangles]
        points2 = [t.point2 for t in self.triangles]
        points3 = [t.point3 for t in self.triangles]
        
        return list(set(points1 + points2 + points3))
    def extract_points_BIS(self, min_distance:float = 0.001):
        points = []
        print(len(self.triangles))
        for i, t in enumerate(self.triangles):
            distance12 = t.point1.point_distance(t.point2)
            distance13 = t.point1.point_distance(t.point3)
            distance23 = t.point2.point_distance(t.point3)
            if distance12 > min_distance:
                n_div = int(distance12 / min_distance)
                for n in range(n_div):
                    new_point = t.point1 + (t.point2 - t.point1)*n/(n_div)
                    points.append(new_point)
            if distance13 > min_distance:
                n_div = int(distance13 / min_distance)
                for n in range(n_div):
                    new_point = t.point1 + (t.point3 - t.point1)*(n+1)/(n_div)
                    points.append(new_point)
            if distance23 > min_distance:
                n_div = int(distance23 / min_distance)
                for n in range(n_div):
                    new_point = t.point2 + (t.point3-t.point2)*n/(n_div)
                    points.append(new_point)
        
        print('all_points available ',len(points))
        valid_points = vm.Vector3D.remove_duplicate(points)
        print('valid points: ', len(valid_points))
        return valid_points
    
    @classmethod
    def from_display_mesh(cls, mesh):
        triangles = []
        for i1, i2, i3 in mesh.triangles:
            triangles.append(vmf.Triangle3D(mesh.points[i1],
                                            mesh.points[i2],
                                            mesh.points[i3]))
        return cls(triangles)


# -*- coding: utf-8 -*-
"""

"""

import math
import numpy as npy


import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D

import volmdlr as vm
import volmdlr.core
import volmdlr.wires as vmw
import volmdlr.faces as vmf
import volmdlr.step as vstep
import volmdlr.stl as vmstl
import volmdlr.edges as vme
import dessia_common as dc

class PointCloud3D(dc.DessiaObject):
    def __init__(self, points, name: str=''):
        self.points = points
        self.name = name
        
    @classmethod
    def from_stl(cls, file_path):
        list_points = vmstl.Stl.from_file(file_path).extract_points_BIS()
        
        return cls(list_points, name='from_stl')
    
    def _bounding_box(self):
        return vm.core.BoundingBox.from_points(self.points)
    
    def to_2d(self, plane_origin, x, y):
        list_points2d = [pt3d.to_2d(plane_origin, x, y) for pt3d in self.points]
        return PointCloud2D(list_points2d, name='3d_to_2d')
    
    def extract(self, u, umin, umax):# -> List[PointCloud3D] :
        extracted_points = []
        for points in self.points :
            dist_to_plane = points.dot(u)
            if dist_to_plane > umin and dist_to_plane < umax :
                extracted_points.append(points)
        return PointCloud3D(extracted_points)
        
    
    def to_shell(self, resolution = 10, normal = None):
        #normal has to be a fondamental vector : X3D, Y3D or Z3D
        bbox = self._bounding_box()
        xyz_bbox = [[bbox.xmin, bbox.xmax], [bbox.ymin,bbox.ymax], [bbox.zmin,bbox.zmax]]
        xyz_list = [l[1]-l[0] for l in xyz_bbox]
        absxyz_list, xyz_vect = [abs(length) for length in xyz_list], [vm.X3D, vm.Y3D, vm.Z3D]

        if normal is None :
            posmax = xyz_list.index(max(absxyz_list))
            normal = xyz_vect[posmax]
            
        else :
            posmax = 0
            for n, vect in enumerate(xyz_vect):
                if vect == normal :
                    posmax = n
        dist_between_plane = xyz_list[posmax]/(resolution-1)
        position_plane = [xyz_bbox[posmax][0] + n*dist_between_plane for n in range(resolution)]
        
        subcloud3d = [self.extract(normal, pos_plane-dist_between_plane/2, pos_plane+dist_between_plane/2) for pos_plane in position_plane]
        print('subcloud3D CREATED')
        vec1, vec2 = xyz_vect[posmax-2], xyz_vect[posmax-1]
        subcloud2d = [subcloud3d[n].to_2d(position_plane[n]*normal, vec1, vec2) for n in range(resolution)]
        # return subcloud2d
        # fig = plt.figure()
        # ax = fig.add_subplot(111)
        # for point in subcloud2d[0].points[1::10]:
            # point.plot(ax=ax)
        print('subcloud2D CREATED')
        print('CREATING POLYGONS')
        initial_polygon2d = [cloud2d.to_polygon() for cloud2d in subcloud2d]
        
        polygon2d, polygon3d = [], []
        for pos_plane, poly in zip(position_plane, initial_polygon2d) :
            if poly is None :
                resolution -= 1
            else :
                polygon2d.append(poly)
                polygon3d.append(poly.to_3d(pos_plane*normal, vec1, vec2))
        # return polygon3d
                
        faces = []
        max_poly_resolution = int(sum([len(poly.points) for poly in polygon3d])/len(polygon3d))+1
        
        fig = plt.figure()
        ax = Axes3D(fig)
        for poly in polygon3d :
            poly.plot(ax=ax)
        
        for n in range(resolution):
            print('sewing polygon', round(n/resolution*100, 2), '%')
            poly1 = polygon3d[n]
            poly1 = poly1.simplify(0.05, 0.1)
            fig = plt.figure()
            ax = fig.add_subplot(111, projection='3d')
            for point in poly1.points:
                point.plot(ax= ax, color = 'g')
            for line in poly1.line_segments:
                line.plot(ax= ax)
            if n == resolution-1 or n == 0:
                plane3d = vmf.Plane3D.from_plane_vectors(position_plane[n]*normal, vec1, vec2)
                surf2d = vmf.Surface2D(polygon2d[n],[])
                faces.append(vmf.PlaneFace3D(plane3d, surf2d))
            if n != resolution-1:
                poly2 = polygon3d[n+1]
                poly2 = poly2.simplify(0.05, 0.1)
                # coords = poly1.sewin1g_with(poly2, vec1, vec2, normal, resolution = max_poly_resolution)
                coords = poly1.sewing(poly2)
                for trio in coords :
                    faces.append(vmf.Triangle3D(trio[0], trio[1], trio[2]))
                # volum = volmdlr.core.VolumeModel(faces)
                # volum.babylonjs()
                # fig = plt.figure()
                # ax = fig.add_subplot(111, projection='3d')
                # poly1.plot(ax=ax, color='g')
                # poly2.plot(ax= ax, color = 'r')
                # for point in poly1.points + poly2.points:
                #     point.plot(ax= ax, color = 'b')
                
        
        return vmf.ClosedShell3D(faces)

    @classmethod        
    def from_step(cls, step_file:str):
        step = vstep.Step(step_file)
        points = step.to_points()
        return cls(points)
    
    def plot(self, ax = None):
        ax = self.points[0].plot(ax = ax)
        for point in self.points[1::1000]:
            point.plot(ax = ax)
            
        return ax
    
class PointCloud2D(dc.DessiaObject):
    def __init__(self, points, name: str=''):
        self.points = points
        self.name = name
        
    def plot(self):
        fig, ax = plt.subplots()
        for pt in self.points :
            pt.plot(ax=ax)
        return ax
    
    def to_polygon(self):
        polygon = vmw.ClosedPolygon2D.points_convex_hull(self.points)
        # polygon = vmw.ClosedPolygon2D.concave_hull(self.points, 0.2, 0.000005)
        # polygon = vmw.ClosedPolygon2D.convex_hull_points(self.points)
        # fig = plt.figure()
        # ax = fig.add_subplot(111)
        # for point in polygon.points:
        #     point.plot(ax= ax, color = 'g')
        # for line in polygon.line_segments:
        #     line.plot(ax= ax)
        # polygon.plot()
        # print('polygon:', polygon)
        if polygon is None or math.isclose(polygon.area(), 0, abs_tol = 1e-6) :
            return None
        else : 
            return polygon